package lc.kra.cell;


import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;
import lc.kra.cell.gui.Provider;

/**
 * Animal World
 *
 * Copyright 2010, 2013 Kristian Kraljic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * @author <a href=mailto:kristian@kra.lc>Kristian Kraljic</a>
 */
public class AnimalWorld extends JComponent implements Provider<AnimalWorld> {
	private static final long serialVersionUID = 1l;
	protected final static Map<Attribute,String> attributes;
	static {
		attributes = new HashMap<Attribute,String>();
		attributes.put(Attribute.NAME,"Animal World");
		attributes.put(Attribute.IDEA,"Lutz Strngmann");
		attributes.put(Attribute.CODE,"Kristian Kraljic");
		attributes.put(Attribute.COPYRIGHT,"(c) 2013 Kristian Kraljic");
		attributes.put(Attribute.DESCRIPTION,"Three kinds of creatures live on a planet in joy and harmony. Only creatures of different races may breed.\n\nIf a creature has found a partner, two children of the third reace are born. Sadly both parents die.\n\nIn case there is only once race of animals left, the planet dies.");
	}
	
	public List<Animal> animals;
	private int age;
	
	public AnimalWorld() {
		animals = Collections.synchronizedList(new ArrayList<Animal>());
		setPreferredSize(new Dimension(300,200));
	}

	@Override public void paintComponent(Graphics defaultGraphics) {
		super.paintComponent(defaultGraphics);
		Graphics2D graphics = (Graphics2D)defaultGraphics;
		graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		graphics.setColor(Color.WHITE);
		graphics.fillRect(0,0,getWidth(),getHeight());
		int count[] = new int[Race.values().length];
		for(Animal animal:animals) {
			count[animal.race.ordinal()]++;
			graphics.setColor(animal.race.color);
			graphics.fillOval(animal.left-5,animal.top-5,10,10);
		}
		graphics.setColor(Color.BLACK);
		graphics.drawString("Age: "+age,10,20);
		for(Race race:Race.values()) {
			graphics.setColor(race.color);
			graphics.drawString(race+": "+count[race.ordinal()],(race.ordinal()+1)*40+30,20);
		}
		graphics.setColor(Color.BLACK);
		graphics.drawString("(Total: "+animals.size()+")",(Race.values().length+1)*40+30,20);
	}
	
	public void extinguish() {
		age = 0;
		animals.clear();
		repaint();
	}
	
	public void populate() {
		try {
			populate(10);
		} catch(NumberFormatException e) { populate(10); }
	}
	public void populate(int count) {
		for(;count>0;count--)
			animals.add(new Animal(Race.values()[(int)(Math.random()*Race.values().length)]));
		repaint();
	}
	
	public void evolve() {
		Animal animal = animals.get((int)(Math.random()*animals.size())),partner = null;
		for(Animal mate:animals)
			if(animal.race!=mate.race) { 
				partner = mate;
				break;
			}
		if(partner==null) return;
		
		animals.remove(animal);
		animals.remove(partner);
		Race birth = null;
		if((animal.race==Race.A&&partner.race==Race.B)
		 ||(animal.race==Race.B&&partner.race==Race.A))
			birth = Race.C;
		if((animal.race==Race.B&&partner.race==Race.C)
		 ||(animal.race==Race.C&&partner.race==Race.B))
			birth = Race.A;
		if((animal.race==Race.A&&partner.race==Race.C)
		 ||(animal.race==Race.C&&partner.race==Race.A))
			birth = Race.B;
		animals.add(new Animal(birth));
		animals.add(new Animal(birth));
		
		age++;
		repaint();
	}
	
	static enum Race {
		A(new Color(255,0,0,196)),B(new Color(0,255,0,196)),C(new Color(0,0,255,196));
		final Color color; private Race(Color color) { this.color = color; }
	}
	
	class Animal {			
		final Race race;
		int left,top;
		Animal(Race race) {
			this.race = race;
			this.left = (int)(Math.random()*(getPreferredSize().width-20))+10;
			this.top = (int)(Math.random()*(getPreferredSize().height-20))+10;
		}
	}

	@Override public Map<Attribute,String> getAttributes() { return attributes; }
	@Override public AnimalWorld getVisualization() { return this; }
	@Override public Component getControls(AnimalWorld visualization) {
		JPanel controls = new JPanel(new FlowLayout(FlowLayout.LEFT,10,0));
		
		JButton extinguish = new JButton("New");
		controls.add(extinguish);
		extinguish.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent e) {
				AnimalWorld.this.extinguish();
				AnimalWorld.this.populate();
			}
		});
		JButton evolve = new JButton("Evolve");
		controls.add(evolve);
		evolve.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent e) {
				AnimalWorld.this.evolve();
			}
		});
		JButton evolveDecade = new JButton("Evolve Decade");
		controls.add(evolveDecade);
		evolveDecade.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent e) {
				for(int decade=10;decade>0;decade--)
					AnimalWorld.this.evolve();
			}
		});
		
		return controls;
	}

	@Override public void setUp() {
		this.extinguish();
		this.populate();
	}

	@Override public void tearDown() {
		this.extinguish();
	}
	
	@Override public String toString() { return attributes.get(Attribute.NAME); }
}
