package lc.kra.cell;


import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import lc.kra.cell.gui.Provider;

/**
 * Game of Life 
 *
 * Copyright 2013 Kristian Kraljic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * @author <a href=mailto:kristian@kra.lc>Kristian Kraljic</a>
 */
public class GameOfLife extends JComponent implements Provider<GameOfLife> {
	private static final long serialVersionUID = 1l;
	protected final static Map<Attribute,String> attributes;
	static {
		attributes = new HashMap<Attribute,String>();
		attributes.put(Attribute.NAME,"Game of Life");
		attributes.put(Attribute.IDEA,"John Horton Conway");
		attributes.put(Attribute.CODE,"Kristian Kraljic");
		attributes.put(Attribute.COPYRIGHT,"(c) 2013 Kristian Kraljic");
		attributes.put(Attribute.DESCRIPTION,"The Game of Life, also known simply as Life, is a cellular automaton devised by the British mathematician John Horton Conway in 1970. \n\nThe \"game\" is a zero-player game, meaning that its evolution is determined by its initial state, requiring no further input. One interacts with the Game of Life by creating an initial configuration and observing how it evolves.");
		attributes.put(Attribute.SOURCE,"http://en.wikipedia.org/wiki/Conway%27s_Game_of_Life");
	}
	
	public final int ROWS = 40,COLUMNS = 60;
	protected static final int CELL = 9;
	
	protected int generation;
	protected boolean[][] cells;
	private Thread simulation;
	
	public GameOfLife() {
		setBackground(Color.WHITE);
		extinguish(); //erase everything
		setPreferredSize(new Dimension(COLUMNS*CELL+3,ROWS*CELL+3));
		MouseAdapter mouse = new MouseAdapter() {
			private int lastRow,lastColumn;
			@Override public void mousePressed(MouseEvent event) {
				int row = event.getY()/CELL,column = event.getX()/CELL;
				lastRow = row; lastColumn = column;
				if(row>=0&&row<ROWS&&column>=0&&column<COLUMNS) {
					cells[row][column] = !cells[row][column];
					repaint(); fireActionPerformed(new ActionEvent(this,3,"set|"+row+"|"+column));
				}
			}
			@Override public void mouseDragged(MouseEvent event) {
				int row = event.getY()/CELL,column = event.getX()/CELL;
				if(row==lastRow&&column==lastColumn) return;
				lastRow = row; lastColumn = column;
				if(row>=0&&row<ROWS&&column>=0&&column<COLUMNS&&!cells[row][column]) {
					cells[row][column] = true;
					repaint(); fireActionPerformed(new ActionEvent(this,3,"set|"+row+"|"+column));
				}
			}
		};
		addMouseListener(mouse);
		addMouseMotionListener(mouse);
	}
	
	public int generation() { return generation; }
	public int countAlive() {
		int count = 0;
		for(int row=0;row<ROWS;row++)
			for(int column=0;column<COLUMNS;column++)
				if(cells[row][column])
					count++;
		return count;
	}
	public int countDead() {
		return ROWS*COLUMNS-countAlive();
	}
	public int countNeighbors(int searchRow,int searchColumn) { //moore neighborhood
		if(searchRow<0||searchRow>=ROWS||searchColumn<0||searchColumn>=COLUMNS)
			return 0;
		int count = cells[searchRow][searchColumn]?-1:0;
		for(int row=searchRow-1;row<=searchRow+1;row++)
			for(int column=searchColumn-1;column<=searchColumn+1;column++)
				if(row>=0&&row<ROWS&&column>=0&&column<COLUMNS&&cells[row][column])
					count++;
		return count;
	}
	
	@Override public void paintComponent(Graphics defaultGraphics) {
		super.paintComponent(defaultGraphics);
		Graphics2D graphics = (Graphics2D)defaultGraphics;
		graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		graphics.setColor(getBackground());
		graphics.fillRect(0,0,getWidth(),getHeight());
		
		graphics.setColor(Color.GRAY);
		for(int row=1;row<=ROWS-1;row++)
			graphics.drawLine(0,row*CELL,getWidth(),row*CELL);
		for(int column=1;column<=COLUMNS-1;column++)
			graphics.drawLine(column*CELL,0,column*CELL,getHeight());
		
		graphics.setColor(Color.BLACK);
		for(int row=0;row<ROWS;row++)
			for(int column=0;column<COLUMNS;column++)
				if(cells[row][column])
					paintCell(graphics,row,column);
	}
	protected void paintCell(Graphics graphics,int row,int column) {
		graphics.fillOval(column*CELL+2,row*CELL+2,CELL-3,CELL-3);
	}

	public void extinguish() {
		generation = 0;
		cells = new boolean[ROWS][COLUMNS];
		repaint(); fireActionPerformed(new ActionEvent(this,1,"extinguish"));
	}
	
	public void compute() {
		boolean[][] cells = new boolean[ROWS][COLUMNS];
		
		for(int row=0;row<ROWS;row++)
			for(int column=0;column<COLUMNS;column++) {
				int neighbors = countNeighbors(row,column);
				switch(neighbors) {
				case 2: cells[row][column] = this.cells[row][column]; break;
				case 3: cells[row][column] = true; break;
				default:cells[row][column] = false; }
			}
		
		generation++;
		this.cells = cells;
		repaint(); fireActionPerformed(new ActionEvent(this,2,"compute"));
	}

  public void addActionListener(ActionListener listener) {
  	listenerList.add(ActionListener.class,listener);
  }
  public void removeActionListener(ActionListener listener) {
	  listenerList.remove(ActionListener.class,listener);
	}
	public ActionListener[] getActionListeners() {
		return (ActionListener[])(listenerList.getListeners(ActionListener.class));
	}
	protected void fireActionPerformed(ActionEvent event) {
		for(ActionListener listener:getActionListeners())
			listener.actionPerformed(event);
	}

	@Override public Map<Attribute,String> getAttributes() { return attributes; }
	@Override public GameOfLife getVisualization() { return this; }
	@Override public Component getControls(final GameOfLife visualization) {		
		JPanel controls = new JPanel(new FlowLayout(FlowLayout.LEFT,10,0));
		
		JButton extinguish = new JButton("Extinguish");
		controls.add(extinguish);
		extinguish.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent actionevent) {
				GameOfLife.this.extinguish();
			}
		});
		JButton next = new JButton("Next Generation");
		controls.add(next);
		next.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent actionevent) {
				GameOfLife.this.compute();
			}
		});
		final JToggleButton simulate = new JToggleButton("Simulate");
		controls.add(simulate);
		simulate.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent arg0) {
				if(simulate.isSelected())
					(simulation = new Thread() {
						@Override public void run() {
							try {
								while(true) {
									GameOfLife.this.compute();
									Thread.sleep(100);
								}
							} catch(InterruptedException e) {}
						}
					}).start();
				else simulation.interrupt();
			}
		});
		final JLabel status = new JLabel();
		controls.add(status);
		GameOfLife.this.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent event) {
				status.setText("Generation: "+GameOfLife.this.generation()+", Alive: "+GameOfLife.this.countAlive());
			}
		});
		GameOfLife.this.getActionListeners()[0].actionPerformed(null);
		
		return controls;
	}

	@Override public void setUp() {
		this.extinguish();
	}

	@Override public void tearDown() {
		this.extinguish();
		if(simulation!=null)
			simulation.interrupt();
	}
	
	@Override public String toString() { return attributes.get(Attribute.NAME); }
}