package lc.kra.cell;


import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.util.HashMap;
import java.util.Map;
import javax.swing.JComponent;
import lc.kra.cell.gui.Provider;

/**
 * Mandelbrot Set
 *
 * Copyright 2010, 2013 Kristian Kraljic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * @author <a href=mailto:kristian@kra.lc>Kristian Kraljic</a>
 */
public class Mandelbrot extends JComponent implements Provider<Mandelbrot> {
	private static final long serialVersionUID = 1l;
	protected static final Map<Attribute,String> attributes;
	static {
		attributes = new HashMap<Attribute,String>();
		attributes.put(Attribute.NAME,"Mandelbrot Set");
		attributes.put(Attribute.IDEA,"Benoit Mandelbrot");
		attributes.put(Attribute.CODE,"Kristian Kraljic");
		attributes.put(Attribute.COPYRIGHT,"(c) 2013 Kristian Kraljic");
		attributes.put(Attribute.DESCRIPTION,"The Mandelbrot set is a mathematical set of points whose boundary is a distinctive and easily recognizable two-dimensional fractal shape. The set is closely related to Julia sets (which include similarly complex shapes), and is named after the mathematician Benoit Mandelbrot, who studied and popularized it.");
		attributes.put(Attribute.SOURCE,"http://en.wikipedia.org/wiki/Mandelbrot_set");
	}
	
	private static final int ITERATIONS = 50,BAND = 3,WIDTH = 500,HEIGHT = 400; 
	
	private int zoom = 0;
	private int[] pixels = new int[WIDTH*HEIGHT*BAND];
	private int[][] colors = new int[256][BAND];
	private BufferedImage image = new BufferedImage(WIDTH,HEIGHT,BufferedImage.TYPE_INT_RGB);
	
	private Complex border = new Complex(),dimension = new Complex();
	
	public Mandelbrot() {
		setBorder(-3.25,-2);
		setDimension(5,4);
		setColor(new Color(10,60,125));
		setBackground(Color.WHITE);
		setPreferredSize(new Dimension(WIDTH,HEIGHT));
		addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent event) {
				double factor;
				if(event.getButton()==MouseEvent.BUTTON1) {
					setBorder(border.real+(dimension.real*((double)event.getX()/WIDTH))-dimension.real/2,
				            border.imaginary+(dimension.imaginary*((double)event.getY()/HEIGHT))-dimension.imaginary/2);
					zoom++; factor = 1d/2;
				} else if(zoom<=1) {
					if(zoom!=0)
						setUp();
					return;
				} else { zoom--; factor = 1d*2; }
				setDimension(dimension.real*factor,dimension.imaginary*factor);
				if(event.getButton()==MouseEvent.BUTTON1)
						 setBorder(border.real+dimension.real*factor,border.imaginary+dimension.imaginary*factor);				
				else setBorder(border.real-dimension.real/(factor*2),border.imaginary-dimension.imaginary/(factor*2));
				repaint();
			}
		});
	}
	
	public void setBorder(double left,double top) {
		this.border.real = left;
		this.border.imaginary = top;
	}
	public void setDimension(double width,double height) {
		this.dimension.real = width;
		this.dimension.imaginary = height;
	}
	public void setColor(Color color) {
		colors[255][0] = colors[255][1] = colors[255][2] = 255;
		int components[] = new int[]{color.getRed(),color.getGreen(),color.getBlue()};
		for(int alpha=0;alpha<=254;alpha++) {
			double opacity = (double)alpha/255;
			for(int band=0;band<BAND;band++)
				colors[alpha][band] = (int)(opacity*components[band]+(1-opacity)*255);
		}
	}
	
	public void paintComponent(final Graphics graphics) {
		Complex step = new Complex(dimension.real/WIDTH,dimension.imaginary/HEIGHT),c = new Complex();
  	c.imaginary = border.imaginary;
  	for(int y=0;y<HEIGHT;y++) {
  		c.real = border.real;
      for(int x=0;x<WIDTH;x++) {
      	int z = BAND*(y*WIDTH+x),iterations = calculate(c),color = (int)(255d/ITERATIONS*iterations);
        if(iterations<ITERATIONS)
        	for(int band=0;band<BAND;band++)
        		 pixels[z+band] = colors[color][band];
        else pixels[z] = pixels[z+1] = pixels[z+2] = 0;
        c.real += step.real;
      }
  		c.imaginary += step.imaginary;
    }
  	
  	WritableRaster raster = image.getRaster();
  	raster.setPixels(0,0,WIDTH,HEIGHT,pixels);
  	graphics.drawImage(image,0,0,null);
  	
  	graphics.setColor(Color.BLACK);
  	graphics.drawString("Click to Zoom In, Right-Click to Zoom Out",5,HEIGHT-10);
    /*graphics.drawLine(getWidth()/2-5,getHeight()/2,getWidth()/2+5,getHeight()/2);
      graphics.drawLine(getWidth()/2,getHeight()/2-5,getWidth()/2,getHeight()/2+5);*/
	}
	public int calculate(Complex c) {
		Complex z_last = new Complex();
		for(int i=0;i<ITERATIONS;i++) {
			Complex z = z_last.pow().add(c);
			if(z.value()>2) return i;
			z_last = z;
		}
		return Integer.MAX_VALUE;
	}
	
	@Override public Map<Attribute,String> getAttributes() { return attributes; }
	@Override public Mandelbrot getVisualization() { return this; }
	@Override public Component getControls(Mandelbrot visualization) { return null; }

	@Override public void setUp() {
		zoom = 0;
		setBorder(-3.25,-2);
		setDimension(5,4);
		repaint();
	}

	@Override public void tearDown() {}
	
	@Override public String toString() { return attributes.get(Attribute.NAME); }
	
	protected static class Complex {
		public double real;
		public double imaginary;
		public Complex() { this(0,0); }
		public Complex(double real,double imaginary) {
			this.real = real;
			this.imaginary = imaginary;
		}
		public Complex subtract(Complex subtrahend) { return add(new Complex(-1*subtrahend.real,-1*subtrahend.imaginary));	}
		public Complex add(Complex summand) {
			Complex sum = new Complex();
			sum.real = this.real+summand.real;
			sum.imaginary = this.imaginary+summand.imaginary;
			return sum;
		}
		public Complex pow() { return multiply(this); }
		public Complex multiply(double factor) {
			Complex product = new Complex(this.real,this.imaginary);
			product.real *= factor;
			product.imaginary *= factor;
			return product;		
		}
		public Complex multiply(Complex factor) {
			Complex product = new Complex();
			product.real = this.real*factor.real-this.imaginary*factor.imaginary;
			product.imaginary = this.real*factor.imaginary+factor.real*this.imaginary;
			return product;
		}
		public double value() {
			return Math.sqrt(this.real*this.real+this.imaginary*this.imaginary);
		}
		public String toString() {
			return new StringBuilder().append('(').append(real).append(',').append(imaginary).append("i)").toString();
		}
	}
}
