package lc.kra.cell.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.font.TextAttribute;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;
import javax.swing.text.JTextComponent;
import lc.kra.cell.AgeingGameOfLife;
import lc.kra.cell.AnimalWorld;
import lc.kra.cell.GameOfLife;
import lc.kra.cell.Mandelbrot;
import lc.kra.cell.WolframUniverse;
import lc.kra.cell.gui.Provider.Attribute;

/**
 * CellularAutomata Suite 
 *
 * Copyright 2013 Kristian Kraljic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * @author <a href=mailto:kristian@kra.lc>Kristian Kraljic</a>
 */
public class Application extends JFrame {
	private static final long serialVersionUID = 1L;
	
	public final List<Provider<?>> providers;
	protected Provider<?> provider;
	protected ProviderContainer container;
	
	public static void main(String[] args) {
		try { UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName()); }
		catch(Exception e_a) {
			try { UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName()); }
			catch(Exception e_b) { /* ... */ }
		} finally { new Application().setVisible(true); }
	}
	
	public Application() {
		providers = new ArrayList<Provider<?>>();	
		for(Class<?> provider:new Class<?>[]{AnimalWorld.class,WolframUniverse.class,GameOfLife.class,AgeingGameOfLife.class,Mandelbrot.class})
			try {
				provider.asSubclass(Provider.class);
				providers.add((Provider<?>)provider.newInstance());
			}	catch(Exception e) { continue; }
		
		Provider<?> provider = null;
		if(providers.size()==0)
			provider = new Provider<Component>() {
				private Map<Attribute,String> attributes;				
				@Override public Map<Attribute,String> getAttributes() {
					if(attributes==null) {
						attributes = new HashMap<Attribute,String>();
						attributes.put(Attribute.NAME,"No Visualization");
						return attributes;
					} else return attributes;
				}
				@Override public Component getVisualization() {
					return (new JPanel() {
						private static final long serialVersionUID = 1l;
						private JPanel build() {
							return this;
						}
					}).build();
				}
				@Override public Component getControls(Component visualization) { return null; }
				@Override public void setUp() {}
				@Override public void tearDown() {}
			};
		else provider = providers.get(0);
		
		final JComboBox<Provider<?>> combo = new JComboBox<Provider<?>>(providers.toArray(new Provider[0]));
		combo.addItemListener(new ItemListener() {
			@Override public void itemStateChanged(ItemEvent event) {
				Application.this.activateProvider(combo.getItemAt(combo.getSelectedIndex()));
			}
		});
		
		((JPanel)this.getContentPane()).setBorder(new EmptyBorder(10,10,10,10));
		this.setLayout(new BorderLayout(10,10));
		this.add(container=new ProviderContainer(),BorderLayout.CENTER);
		this.add(combo,BorderLayout.NORTH);
		this.activateProvider(provider);
		this.setLocationByPlatform(true);
		this.setDefaultCloseOperation(EXIT_ON_CLOSE);
		this.setResizable(false);
	}
	
	public <Visualization extends Component> void activateProvider(final Provider<Visualization> provider) {
		Map<Attribute,String> attributes = provider.getAttributes();
		if(attributes==null) attributes = Collections.emptyMap();
		
		StringBuilder title = new StringBuilder("CellularAutomata Suite").append(" - ");
		if(attributes.containsKey(Attribute.NAME))
			   title.append(attributes.get(Attribute.NAME));
		else title.append(provider.getClass().getSimpleName());
		if(attributes.containsKey(Attribute.IDEA)||attributes.containsKey(Attribute.CODE)) {
			title.append(" (");
			String idea = attributes.get(Attribute.IDEA),code = attributes.get(Attribute.CODE);
			if(idea!=null) {
				if(!idea.equals(code)) {
					title.append("Idea: ").append(idea);
					if(code!=null)
						title.append(", ");
				} else title.append("Idea & ");
			}
			if(code!=null)
				title.append("Implementation: ").append(code);
			title.append(")");
		}
		this.setTitle(title.toString());
		
		if(this.provider!=null)
			this.provider.tearDown();
		this.container.setVisualization(provider);
		(this.provider=provider).setUp();
	}
	
	protected class ProviderContainer extends JPanel {
		private static final long serialVersionUID = 1l;
		protected Provider<?> provider;
		
		private JScrollPane container;
		private Component visualization,controls;
		private JTextComponent description;
		private JLabel source;
		
		public ProviderContainer() {
			super(new BorderLayout(10,10));
			JPanel pane = new JPanel(new BorderLayout(0,6));
			pane.add(description=(new JTextArea() {
				private static final long serialVersionUID = 1l;
				private JTextComponent build() {
	        setEditable(false);
	        setCursor(null);
	        setOpaque(false);
	        setFocusable(false);
	        setLineWrap(true);
	        setWrapStyleWord(true);
	        setFont(UIManager.getFont("Label.font").deriveFont(Font.PLAIN));
					return this;
				}
			}).build(),BorderLayout.NORTH);
			source = (new JLabel() {
				private static final long serialVersionUID = 1l;
				private JLabel build() {
	        setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
	        setOpaque(false);
	        setFocusable(false);
	        Map<TextAttribute,Object>  attributes = new HashMap<TextAttribute, Object>();
	        attributes.put(TextAttribute.UNDERLINE,TextAttribute.UNDERLINE_ON);
	        setFont(UIManager.getFont("Label.font").deriveFont(Font.PLAIN).deriveFont(attributes));
	        setForeground(Color.BLUE);
	        addMouseListener(new MouseAdapter() {
						@Override public void mouseClicked(MouseEvent event) {
							if(!getText().isEmpty())
								try { Desktop.getDesktop().browse(URI.create(getText())); }
								catch(Exception e) { /* ... */ }
						}
					});
					return this;
				}
			}).build();
			this.add(pane,BorderLayout.NORTH);
			this.add(container=new JScrollPane(),BorderLayout.CENTER);
		}
		
		public <Visualization extends Component> void setVisualization(Provider<Visualization> provider) {
			Map<Attribute,String> attributes = provider.getAttributes();
			
			description.setText(null); source.setText(null);
			description.setSize(new Dimension(0,0));
			
			Visualization visualization = provider.getVisualization();
			if(this.visualization!=null) container.setViewportView(null);
			container.setViewportView(this.visualization=visualization);
			Dimension size = visualization.getPreferredSize();
			container.setHorizontalScrollBarPolicy(size.width>800?JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS:JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
			container.setVerticalScrollBarPolicy(size.height>600?JScrollPane.VERTICAL_SCROLLBAR_ALWAYS:JScrollPane.VERTICAL_SCROLLBAR_NEVER);
			size.width = Math.max(200,Math.min(size.width,800));		
			size.height = Math.max(200,Math.min(size.height,600));
			container.setPreferredSize(size);
			
			if(this.controls!=null) this.remove(this.controls);
			if((this.controls=provider.getControls(visualization))!=null)
				this.add(controls,BorderLayout.SOUTH);
			Application.this.pack();
			
			description.setVisible(true);
			if(attributes.containsKey(Attribute.DESCRIPTION))
			     description.setText(attributes.get(Attribute.DESCRIPTION));
			else description.setVisible(false);
			
			description.getParent().remove(source);
			if(attributes.containsKey(Attribute.SOURCE)) {
				description.getParent().add(source,BorderLayout.SOUTH);
				source.setText(attributes.get(Attribute.SOURCE));
			}

			Application.this.pack();
		}
	}
}